// $Id: IterEventState.java 248 2005-08-25 18:16:27Z davidb $
//
// Copyright (C) 2005 VeriSign, Inc.

package se.rfc.unbound.iter;

import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;

import se.rfc.unbound.*;

public class IterEventState implements ConcurrentRequestState
{
  /**
   * Externally generated queries start at this state. Query restarts are
   * reset to this state.
   */
  protected static final int INIT_REQUEST_STATE   = 0;

  /**
   * Root priming events reactivate here, most other events pass through this
   * naturally as the 2nd part of the INIT_REQUEST_STATE.
   */
  protected static final int INIT_REQUEST_2_STATE = 1;

  /**
   * Stub priming events reactivate here, most other events pass through this
   * naturally as the 3rd part of the INIT_REQUEST_STATE.
   */
  protected static final int INIT_REQUEST_3_STATE = 2;

  /**
   * Each time a delegation point changes for a given query or a query times
   * out and/or wakes up, this state is (re)visited. This state is reponsible
   * for iterating through a list of nameserver targets.
   */
  protected static final int QUERYTARGETS_STATE   = 3;

  /**
   * Responses to queries start at this state. This state handles the decision
   * tree associated with handling responses.
   */
  protected static final int QUERY_RESP_STATE     = 4;

  /** Responses to priming queries finish at this state. */
  protected static final int PRIME_RESP_STATE     = 5;

  /** Responses to target queries start at this state. */
  protected static final int TARGET_RESP_STATE    = 6;

  /** Responses that are to be returned upstream end at this state. */
  protected static final int FINISHED_STATE       = 7;

  /**
   * This is the state that event is in or should sent to -- all requests
   * should start with the INIT_REQUEST_STATE. All responses should start with
   * QUERY_RESP_STATE. Subsequent processing of the event will change this
   * state.
   */
  protected int              state;

  /**
   * This is the state that responses should be routed to once the response is
   * final. For externally initiated queries, this will be FINISHED_STATE,
   * locally initiated queries will have different final states.
   */
  protected int              finalState;

  /**
   * This is a list of (S)RRsets that must be prepended to the ANSWER section
   * of a response before being sent upstream.
   */
  protected List             prependList;
  /**
   * This is the current delegation point for an in-progress query. This
   * object retains state as to which delegation targets need to be
   * (sub)queried for vs which ones have already been visited.
   */
  protected DelegationPoint  currentDelegationPoint;
  /**
   * This is the referral message for the current delegation point. It may be
   * null if the currentDelegationPoint wasn't set via a referral message.
   */
  protected SMessage         currentDelegationMessage;
  /**
   * This is a queue of targets fetched from the current delegation point.
   * Because the DelegationPoint object currently hands back groups of
   * resolved targets, this queue is used to hold the current set of unvisited
   * targets so that they can be dequeued one at a time.
   * 
   * FIXME: this functionality should be rolled back into the DelegationPoint
   * class.
   */
  protected LinkedList       targetQueue;
  /**
   * This is a number representing the number of outstanding target sub
   * queries in progress for this request. Less than zero is used to indicate
   * that the queries have not yet been made.
   */
  protected int              targetQueries;

  /**
   * This is a number representing the number of outstanding direct queries in
   * progress for this request.
   */
  protected int              currentQueries;

  /**
   * This is a number representing the number of times this query has been
   * restarted.
   */
  protected int              queryRestartCount;

  /**
   * This is a number representing the number of times this query as followed
   * a referral.
   */
  protected int              referralCount;
  /**
   * This is flag that, if true, means that this event is representing a stub
   * priming query. It is meaningless unless the finalState is the
   * PRIMING_RESP_STATE.
   */
  protected boolean          primingStub;

  /**
   * A list of concurrent requests that may be attached to this event. If
   * null, then there are no attached requests.
   */
  private List               mConcurrentRequestList;

  public IterEventState()
  {
    // default our targetQueries counter.
    targetQueries = -1;
  }

  /** Reset this state to a new delegation point. */
  public void reset(SMessage delegationMessage, DelegationPoint dp)
  {
    currentDelegationMessage = delegationMessage;
    currentDelegationPoint = dp;
    targetQueries = -1;
    if (targetQueue != null) targetQueue.clear();
    currentQueries = 0;
  }

  public static String stateString(int state)
  {
    switch (state)
    {
      case INIT_REQUEST_STATE :
        return "INIT REQUEST STATE";
      case INIT_REQUEST_2_STATE :
        return "INIT REQUEST STATE (stage 2)";
      case INIT_REQUEST_3_STATE:
        return "INIT REQUEST STATE (stage 3)";
      case QUERYTARGETS_STATE :
        return "QUERY TARGETS STATE";
      case PRIME_RESP_STATE :
        return "PRIME RESPONSE STATE";
      case QUERY_RESP_STATE :
        return "QUERY RESPONSE STATE";
      case TARGET_RESP_STATE :
        return "TARGET RESPONSE STATE";
      case FINISHED_STATE :
        return "FINISHED RESPONSE STATE";
      default :
        return "UNKNOWN ITER STATE " + state;
    }
  }

  public void addConcurrentEvent(DNSEvent event)
  {
    if (mConcurrentRequestList == null)
    {
      mConcurrentRequestList = new ArrayList();
    }
    mConcurrentRequestList.add(event);
  }

  public boolean hasConcurrentEvents()
  {
    if (mConcurrentRequestList != null)
      return !mConcurrentRequestList.isEmpty();
    return false;
  }

  public List getConcurrentEvents()
  {
    return mConcurrentRequestList;
  }

  public void clearConcurrentEvents()
  {
    if (mConcurrentRequestList != null) mConcurrentRequestList.clear();
  }

  public static boolean isResponseState(int nextState)
  {
    switch (nextState)
    {
      case INIT_REQUEST_STATE :
      case INIT_REQUEST_2_STATE :
      case QUERYTARGETS_STATE :
        return false;
      default :
        return true;
    }
  }
}
